# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/pdf/01_extract.ipynb.

# %% auto 0
__all__ = ['add_other_class', 'print_segments', 'text_to_segments', 'text_to_ems', 'load_segs_model', 'load_ems_model',
           'write_segments', 'write_embeddings']

# %% ../../nbs/pdf/01_extract.ipynb 2
from ..imports import *
from .core import *

# %% ../../nbs/pdf/01_extract.ipynb 4
def add_other_class(classes, other_class="Other"):
    # Add 'Other' class if not present
    if other_class not in classes:
        classes.append(other_class)
    return classes


def print_segments(segments: dict, limit: int = 10, width: int = 100):
    for k, v in segments.items():
        print(f"{k}: {len(v)}")
        for s in v[:limit]:
            print("\t", end="")
            pprint(s, width=width)
            print()
        print("-" * (width + 4))


def text_to_segments(
    text,
    segs_model,
    thresh=0.6,
    classes=["Work Experience", "Education", "Certifications", "Other"],
    other_class="Other",
):
    classes = add_other_class(classes, other_class)
    ot_id = classes.index(other_class)
    preds = segs_model.predict(text).detach().cpu().numpy()
    probs = segs_model.predict_proba(text).detach().cpu().numpy().max(1)
    preds[probs < thresh] = ot_id
    pred_classes = [classes[p] for p in preds]
    segments = {pc: [] for pc in classes}
    for pc, txt in zip(pred_classes, text):
        if txt not in segments[pc]:
            segments[pc].append(txt)
    return segments, pred_classes, probs


def text_to_ems(
    df,
    ems_model,
    classes=["Work Experience", "Education", "Certifications", "Other"],
    other_class="Other",
):
    classes = add_other_class(classes, other_class)
    seg_ids = {c: i + 1 for i, c in enumerate(classes)}
    fn = Path(df["path"]).stem
    ems_folder = Path(df["ems_folder"])
    os.makedirs(ems_folder, exist_ok=True)
    ems_path = str(ems_folder / fn)
    txt = df["text"]
    pred_classes = df.get("classes", [""] * len(txt))
    if len(txt) == 0 or len(pred_classes) == 0:
        return df
    ems = ems_model.encode(txt).tolist()
    for i, data in enumerate(zip(pred_classes, ems)):
        pc, em = data
        seg_id = seg_ids.get(pc, 0)
        if seg_id == 0:
            seg = ""
        else:
            seg = f"_{seg_id}"
        jf = ems_path + f"{seg}_{i+1}.json"
        em_dict = {"id": fn, "embedding": em}
        with open(jf, "w") as f:
            json.dump(em_dict, f)
    return df


def load_segs_model(model_name="HamzaFarhan/PDFSegs", device="cpu"):
    return SetFitModel.from_pretrained(model_name).to(device)


def load_ems_model(model_name="HamzaFarhan/PDFSegs", device="cpu"):
    return SentenceTransformer(model_name, device=device)


def write_segments(
    segs_model,
    data_path,
    output_path="pdf_segments",
    n_lines=3,
    classes=["Work Experience", "Education", "Certifications", "Other"],
    other_class="Other",
):
    """
    Extracts text from PDFs and writes segments to JSON files.

    Parameters
    ----------
    segs_model : SetFitModel
        SetFit model for segment classification.
    data_path : Union[str, Path, List[Union[str, Path]]]
        Path to PDFs. Can be a single file, a directory, or a list of files/directories.
    output_path : Union[str, Path]
        Folder to write JSON files. Defaults to 'pdf_segments'.
    n_lines : int
        Number of lines to group together when extracting text from PDFs.
    classes : List[str]
        List of segment classes. Defaults to ["Work Experience", "Education", "Certifications", "Other"].
    other_class : str
        Name of the 'Other' class. Defaults to "Other".
    Returns
    -------
    None
    """
    output_path = Path(output_path)
    os.makedirs(output_path, exist_ok=True)
    pdfs = resolve_data_path(data_path)
    text_dict = extract_text_dict(pdfs, n_lines=n_lines)
    for file in pdfs:
        try:
            pdf_text = text_dict[str(file)]
            (segments,) = text_to_segments(
                pdf_text,
                segs_model,
                thresh=0.6,
                classes=classes,
                other_class=other_class,
            )[0]
            fn = (output_path / Path(file).stem).with_suffix(".json")
            with open(fn, "w") as f:
                json.dump(segments, f, indent=4)
        except Exception as e:
            msg.fail(f"\nCould not write segments for file: {str(file)}", e)
            # print(f"\nCould not write segments for file: {str(file)}\n{e}")


def write_embeddings(
    segs_model,
    ems_model,
    data_path,
    output_path="pdf_ems",
    n_lines=3,
    classes=["Work Experience", "Education", "Certifications", "Other"],
    other_class="Other",
):
    """
    Extracts text from PDFs and writes embeddings to JSON files.

    Parameters
    ----------
    segs_model : SetFitModel
        SetFit model for segment classification.
    ems_model : SentenceTransformer
        SentenceTransformer model for embedding generation.
    data_path : Union[str, Path, List[Union[str, Path]]]
        Path to PDFs. Can be a single file, a directory, or a list of files/directories.
    output_path : Union[str, Path]
        Folder to write JSON files. Defaults to 'pdf_ems'.
    n_lines : int
        Number of lines to group together when extracting text from PDFs.
    classes : List[str]
        List of segment classes. Defaults to ["Work Experience", "Education", "Certifications", "Other"].
    other_class : str
        Name of the 'Other' class. Defaults to "Other".
    Returns
    -------
    None
    """
    os.makedirs(output_path, exist_ok=True)
    classes = add_other_class(classes, other_class)
    seg_ids = {c: i + 1 for i, c in enumerate(classes)}
    # seg_ids = {"Work Experience": 1, "Education": 2, "Certifications": 3, "Other": 4}
    pdfs = resolve_data_path(data_path)
    text_dict = extract_text_dict(pdfs, n_lines=n_lines)
    for file in pdfs:
        # try:
        fn = Path(file).stem
        pdf_text = text_dict[str(file)]
        ems = ems_model.encode(pdf_text)
        pred_classes = text_to_segments(pdf_text, segs_model, classes=classes)[1]
        for i, data in enumerate(zip(pred_classes, ems)):
            pc, em = data
            seg_id = seg_ids[pc]
            jf = Path(output_path) / f"{fn}_{seg_id}_{i+1}.json"
            em_dict = {"id": fn, "embedding": em.tolist()}
            with open(jf, "w") as f:
                json.dump(em_dict, f)
            # except Exception as e:
            # print(f"\nCould not write embeddings for {str(file)} - {e}")
            # msg.fail(f"\nCould not write embeddings for {str(file)}", e)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/pdf/05_app.ipynb.

# %% auto 0
__all__ = ['SegsPredictor', 'NERPredictor', 'EmsPredictor', 'pdf_pipeline', 'ner_pipeline', 'pdf_pipeline_params',
           'ner_pipeline_params']

# %% ../../nbs/pdf/05_app.ipynb 2
from ..imports import *
from ..utils import *
from .core import *
from .extract import *
from .ner import *
from .df import *
from .. import redis_kv_store


# %% ../../nbs/pdf/05_app.ipynb 4
class SegsPredictor:
    def __init__(
        self,
        model_name="HamzaFarhan/PDFSegs",
        thresh=0.6,
        write=False,
        device=None,
        **kwargs,
    ):
        device = default_device() if device is None else device
        self.thresh = thresh
        self.model = load_segs_model(model_name, device=device)
        self.write = write

    def __call__(self, df):
        msg.info(f"LEN SEGS DF: {len(df)}", spaced=True)
        df[["segs", "classes", "probs"]] = df.apply(
            lambda x: write_df_segs(
                x,
                self.model,
                thresh=self.thresh,
                write=self.write,
            ),
            axis=1,
            result_type="expand",
        )
        return df


class NERPredictor:
    def __init__(
        self,
        ner_roles=True,
        device=None,
        task_id=None,
        redis_host="127.0.0.1",
        redis_port=6379,
    ):
        device = default_device() if device is None else device
        # msg.info(f"NER DEVICE: {device}", spaced=True)
        self.tner = load_ner_model(device=device)
        self.jner = load_job_model(device=device)
        if ner_roles:
            self.work_ner_dict = {"company": "", "role": "", "date": ""}
        else:
            self.work_ner_dict = {"company": "", "date": ""}
        self.task_id = task_id
        self.kv_store = redis_kv_store.KeyValueStore(
            redis_host=redis_host, redis_port=redis_port
        )

    def __call__(self, df):
        msg.info(f"LEN NER DF: {len(df)}", spaced=True)
        df = df.apply(
            lambda x: write_df_ner(
                x,
                tner=self.tner,
                jner=self.jner,
                work_ner_dict=self.work_ner_dict,
            ),
            axis=1,
        )
        if self.task_id is not None and self.kv_store is not None:
            for _ in range(len(df)):
                update_task_progress(self.task_id, self.kv_store)
        return df


class EmsPredictor:
    def __init__(
        self,
        model_name="HamzaFarhan/PDFSegs",
        device="cpu",
        task_id=None,
        redis_host="127.0.0.1",
        redis_port=6379,
        **kwargs,
    ):
        device = default_device() if device is None else device
        self.model = load_ems_model(model_name, device=device)
        self.task_id = task_id
        self.kv_store = redis_kv_store.KeyValueStore(
            redis_host=redis_host, redis_port=redis_port
        )

    def __call__(self, df):
        msg.info(f"LEN EMS DF: {len(df)}", spaced=True)
        df = df.apply(lambda x: write_df_ems(x, self.model), axis=1)
        if self.task_id is not None and self.kv_store is not None:
            for _ in range(len(df)):
                update_task_progress(self.task_id, self.kv_store)
        return df


def pdf_pipeline(params):
    try:
        task_id = params["task_id"]
        data_path = params["data_path"]
        segs_folder = params["segs_folder"]
        ems_folder = params["ems_folder"]
        segs_params = params["segs_params"]
        ems_params = params["ems_params"]
        num_blocks = params["num_blocks"]
        blocks_per_window = params["blocks_per_window"]
        redis_host = params.get("redis_host", "127.0.0.1")
        redis_port = params.get("redis_port", 6379)
        do_segs = params.get("do_segs", True)
        do_ner = params.get("do_ner", False)
        if do_segs:
            segs_params["fn_constructor_kwargs"]["write"] = True
    except Exception as e:
        raise Exception(f"Error in params: {e}.\nParams: {params}")
    try:
        kv_store = redis_kv_store.KeyValueStore(redis_host=redis_host, redis_port=redis_port)
        ems_params["fn_constructor_kwargs"]["task_id"] = task_id
    except Exception as e:
        raise Exception(f"Error in connecting to redis: {e}.\nParams: {params}")

    try:
        # segs_folder = Path(task_id) / Path(segs_folder).name
        # ems_folder = Path(task_id) / Path(ems_folder).name
        task_folder = Path(f"/tmp/{task_id}")
        os.makedirs(task_folder, exist_ok=True)
        local_segs_folder = get_local_path(segs_folder, task_folder)
        local_ems_folder = get_local_path(ems_folder, task_folder)
        df = create_pdf_df(
            data_path, segs_folder=local_segs_folder, ems_folder=local_ems_folder
        )
        init_task_progress(task_id, kv_store, len(df))
        ds = rd.from_modin(df)
        if num_blocks is not None:
            num_blocks = min(num_blocks, len(df))
            ds = ds.repartition(num_blocks)
            if blocks_per_window is not None:
                ds = ds.window(blocks_per_window=blocks_per_window)
    except Exception as e:
        raise Exception(f"Error in create_pdf_df: {e}.")
    try:
        ds = ds.map_batches(extract_df_text, batch_size=params["segs_params"]["batch_size"])
    except Exception as e:
        raise Exception(f"Error in extract_df_text: {e}.")

    if do_segs:
        try:
            ds = ds.map_batches(
                SegsPredictor,
                **segs_params,
            )
        except Exception as e:
            raise Exception(f"Error in SegsPredictor: {e}.\nseg_params: {segs_params}")
    try:
        ds = ds.map_batches(
            EmsPredictor,
            **ems_params,
        )
    except Exception as e:
        raise Exception(f"Error in EmsPredictor: {e}.\nems_params: {ems_params}")
    ds = str(write_ds(ds, task_folder / "pdf_preds.parquet"))
    if is_bucket(segs_folder):
        # bucket = Path(segs_folder).parent
        bucket_up(local_segs_folder, segs_folder, only_new=False)
    if is_bucket(ems_folder):
        # bucket = Path(ems_folder).parent
        bucket_move(local_ems_folder, ems_folder)
    shutil.rmtree(ds)
    if not do_ner:
        shutil.rmtree(task_folder)
    return {"pipeline_result": "SUCCESS"}


def ner_pipeline(params):
    try:
        task_id_ner = params["task_id_ner"]
        task_id = params["task_id"]
        data_path = params["data_path"]
        segs_folder = params["segs_folder"]
        ner_params = params["ner_params"]
        num_blocks = params["num_blocks"]
        redis_host = params.get("redis_host", "127.0.0.1")
        redis_port = params.get("redis_port", 6379)
    except Exception as e:
        raise Exception(f"Error in params: {e}.\nParams: {params}")
    try:
        kv_store = redis_kv_store.KeyValueStore(redis_host=redis_host, redis_port=redis_port)
        ner_params["fn_constructor_kwargs"]["task_id"] = task_id_ner
    except Exception as e:
        raise Exception(f"Error in connecting to redis: {e}.\nParams: {params}")
    try:
        task_folder = Path(f"/tmp/{task_id}")
        os.makedirs(task_folder, exist_ok=True)
        local_segs_folder = get_local_path(segs_folder, task_folder)
        # msg.info(f"Local_segs_folder: {local_segs_folder}", spaced=True)
        df = create_ner_df(data_path, segs_folder=local_segs_folder)
        init_task_progress(task_id_ner, kv_store, len(df))
        ds = rd.from_modin(df)
        if num_blocks is not None:
            num_blocks = min(num_blocks, len(df))
            ds = ds.repartition(num_blocks)
    except Exception as e:
        raise Exception(f"Error in create_ner_df: {e}.")

    try:
        ds = ds.map_batches(
            NERPredictor,
            **ner_params,
        )
    except Exception as e:
        raise Exception(f"Error in NERPredictor: {e}.\nner_params: {ner_params}")
    ds = str(write_ds(ds, task_folder / "ner_preds.parquet"))
    if is_bucket(segs_folder):
        bucket_up(local_segs_folder, segs_folder, only_new=True)
    shutil.rmtree(task_folder)
    return {"ner_pipeline_result": "SUCCESS"}


def pdf_pipeline_params(
    data_path="",
    segs_folder="pdf_segs",
    ems_folder="pdf_ems",
    segs_model="HamzaFarhan/PDFSegs",
    ems_model="HamzaFarhan/PDFSegs",
    num_gpus=1,
    num_blocks=None,
    batch_sizes=[8, 8, 8],
    workers=[[3, 4], [3, 4], [3, 4]],
    do_segs=True,
    do_ner=False,
    redis_host="127.0.0.1",
    redis_port=6379,
    **kwargs,
):
    if not is_list(batch_sizes):
        batch_sizes = [batch_sizes]
    if not is_list(workers[0]):
        workers = [workers]
    batch_sizes += [batch_sizes[0]] * (2 - len(batch_sizes))
    workers += [workers[0]] * (2 - len(workers))
    try:
        max_workers = workers[1][1]
        if do_segs:
            max_workers += workers[0][1]
            # if do_ner:
            # max_workers += workers[1][1]
        num_gpus = num_gpus / max_workers
    except Exception as e:
        raise Exception(f"Invalid workers: {e}")
    try:
        params = {
            "data_path": data_path,
            "segs_folder": segs_folder,
            "ems_folder": ems_folder,
            "num_blocks": num_blocks,
            "blocks_per_window": None,
            "do_segs": do_segs,
            "do_ner": do_ner,
            "redis_host": redis_host,
            "redis_port": redis_port,
            "segs_params": dict(
                fn_constructor_kwargs=dict(
                    model_name=segs_model,
                    device=None,
                ),
                batch_size=batch_sizes[0],
                compute=rd.ActorPoolStrategy(min_size=workers[0][0], max_size=workers[0][1]),
                num_gpus=num_gpus,
            ),
            "ems_params": dict(
                fn_constructor_kwargs=dict(
                    model_name=ems_model,
                    device=None,
                    redis_host=redis_host,
                    redis_port=redis_port,
                ),
                batch_size=batch_sizes[1],
                compute=rd.ActorPoolStrategy(min_size=workers[1][0], max_size=workers[1][1]),
                num_gpus=num_gpus,
            ),
        }
    except Exception as e:
        raise Exception(f"Invalid params: {e}.\nParams so far: {locals()}")
    return params


def ner_pipeline_params(
    data_path="",
    segs_folder="pdf_segs",
    num_gpus=1,
    num_blocks=None,
    ner_batch_size=8,
    ner_workers=[[3, 4]],
    redis_host="127.0.0.1",
    redis_port=6379,
    # do_ner=True,
    **kwargs,
):
    try:
        if is_list(ner_batch_size):
            ner_batch_size = ner_batch_size[0]
        if not is_list(ner_workers[0]):
            ner_workers = [ner_workers]
        min_workers, max_workers = ner_workers[0]
        num_gpus = num_gpus / max_workers
    except Exception as e:
        raise Exception(f"Invalid workers: {e}")
    try:
        params = {
            "data_path": data_path,
            "segs_folder": segs_folder,
            "num_blocks": num_blocks,
            "blocks_per_window": None,
            "redis_host": redis_host,
            "redis_port": redis_port,
            "ner_params": dict(
                fn_constructor_kwargs=dict(
                    device=None,
                    redis_host=redis_host,
                    redis_port=redis_port,
                ),
                batch_size=ner_batch_size,
                compute=rd.ActorPoolStrategy(min_size=min_workers, max_size=max_workers),
                num_gpus=num_gpus,
            ),
        }
    except Exception as e:
        raise Exception(f"Invalid params: {e}.\nParams so far: {locals()}")
    return params



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/pdf/04_mappers.ipynb.

# %% auto 0
__all__ = ['SegsMapper']

# %% ../../nbs/pdf/04_mappers.ipynb 2
from ..imports import *
from ..utils import *
from ..mapper import *
from .extract import *
from .df import *

# %% ../../nbs/pdf/04_mappers.ipynb 4
class SegsMapper(Mapper):
    """
    A custom mapper for PDF segmentation. It uses the `df_segs` function to extract the segments from the PDF.
    """

    def __init__(
        self,
        segs_model="HamzaFarhan/PDFSegs",  # The name of the model to use for segmentation.
        udf=df_segs,  # The function to use for extracting the segments.
        udf_kwargs=dict(  # The keyword arguments to pass to the `udf`.
            thresh=0.6,
            classes=["Work Experience", "Education", "Certifications", "Other"],
            other_class="Other",
        ),
        **kwargs,
    ):
        segs_model = load_segs_model(segs_model, device=default_device())
        udf_kwargs["segs_model"] = segs_model
        super().__init__(**locals_to_params(locals()))



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/pdf/03_ner.ipynb.

# %% auto 0
__all__ = ['load_ner_model', 'load_job_model', 'proc_ner', 'job_ner', 'edu_ner', 'work_ner', 'is_valid_jner', 'is_valid_tner']

# %% ../../nbs/pdf/03_ner.ipynb 2
from ..imports import *

# %% ../../nbs/pdf/03_ner.ipynb 4
def load_ner_model(model_name="tner/deberta-v3-large-ontonotes5", device="cpu"):
    tokenizer = AutoTokenizer.from_pretrained(model_name)
    model = AutoModelForTokenClassification.from_pretrained(model_name)
    return pipeline(
        "ner",
        model=model,
        tokenizer=tokenizer,
        aggregation_strategy="simple",
        device=device,
    )


def load_job_model(model_name="ismail-lucifer011/autotrain-job_all-903929564", device="cpu"):
    tokenizer = AutoTokenizer.from_pretrained(model_name)
    model = AutoModelForTokenClassification.from_pretrained(model_name)
    return pipeline(
        "ner",
        model=model,
        tokenizer=tokenizer,
        aggregation_strategy="simple",
        device=device,
    )


def proc_ner(txt, ner, ner_dict={"institute": "", "date": ""}, thresh=3):
    ner_dict = copy.deepcopy(ner_dict)
    org_key = "institute" if "institute" in ner_dict else "company"
    mapper = {
        "ORG": org_key,
        "FAC": org_key,
        "GPE": org_key,
        "LOC": org_key,
        "Job": "role",
        "WORK_OF_ART": "degree",
        "DATE": "date",
    }
    ner_dict["text"] = txt
    for d in ner:
        eg = d["entity_group"]
        w = " " + d["word"].strip()
        k = mapper.get(eg, None)
        if k is not None and ner_dict.get(k, None) is not None and not w.startswith("##"):
            ner_dict[k] = (ner_dict[k] + w).strip()
    res = {k: v for k, v in ner_dict.items() if len(v) > thresh}
    if res.get(org_key, None) is None:
        return {}
    return res


def job_ner(txt, tner, jner):
    return tner(txt) + jner(txt)


def edu_ner(txt, tner, ner_dict={"institute": "", "date": ""}):
    ner = tner(txt)
    return proc_ner(txt, ner, ner_dict)


def work_ner(txt, tner, jner, ner_dict={"company": "", "date": ""}):
    ner = job_ner(txt, tner, jner)
    return proc_ner(txt, ner, ner_dict)


def is_valid_jner(ner, thresh=3):
    return ner.get("company", None) is not None


def is_valid_tner(ner, thresh=3):
    return ner.get("institute", None) is not None



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_utils.ipynb.

# %% auto 0
__all__ = ['json_file', 'get_task_from_kv_store', 'init_task_progress', 'update_task_progress', 'is_bucket', 'gsutil_bucket',
           'gsutil_src', 'bucket_move', 'bucket_up', 'bucket_dl', 'get_local_path', 'lit_eval', 'find_alternate_path',
           'resolve_ds_path', 'write_ds', 'chain_models', 'is_preprocessor', 'chain_processors', 'handle_processors',
           'repartition_ds', 'transform_ds', 'group_df_on']

# %% ../nbs/00_utils.ipynb 2
from dreamai.core import *
from .imports import *

# %% ../nbs/00_utils.ipynb 4
def json_file(path, folder):
    path = Path(path)
    folder = Path(folder)
    os.makedirs(folder, exist_ok=True)
    return folder / f"{path.stem}.json"


def get_task_from_kv_store(task_id, kv_store):
    task = kv_store.get(task_id)
    if task is None:
        raise Exception(f"No task entry found for task_id {task_id}.")
    if type(task) != dict:
        raise Exception(f"Wrong type for task with task_id {task_id}.")
    if len(task) == 0:
        raise Exception(f"Empty dict for task_id {task_id}.")
    return task


def init_task_progress(task_id, kv_store, total):
    task = get_task_from_kv_store(task_id, kv_store)
    task["progress"] = f"processing..."
    task["total"] = total
    kv_store.insert(task_id, task)


def update_task_progress(task_id, kv_store, **kwargs):
    task = get_task_from_kv_store(task_id, kv_store)
    prog = task["progress"]
    total = task["total"]
    if prog == "processing...":
        task["progress"] = f"1/{total}"
    else:
        curr, total = prog.split("/")
        task["progress"] = f"{int(curr) + 1}/{total}"
    kv_store.insert(task_id, task)


def json_file(path, folder):
    path = Path(path)
    folder = Path(folder)
    os.makedirs(folder, exist_ok=True)
    return folder / f"{path.stem}.json"


def is_bucket(p):
    return str(p).startswith("gs://")


def gsutil_bucket(bucket):
    if not str(bucket).startswith("gs://"):
        bucket = "gs://" + str(bucket)
    return bucket


def gsutil_src(folder):
    if Path(folder).suffix != "":
        return str(folder)
    folder = str(folder)
    if folder[-1] != "/":
        folder += "/"
    folder += "*"
    return folder


def bucket_move(folder, bucket):
    gu = shutil.which("gsutil")
    bucket = gsutil_bucket(bucket)
    folder = gsutil_src(folder)
    subprocess.run([gu, "-m", "mv", folder, bucket])


def bucket_up(folder, bucket, only_new=True):
    gu = shutil.which("gsutil")
    bucket = gsutil_bucket(bucket)
    folder = gsutil_src(folder)
    cmd = [gu, "-m", "cp", "-r"]
    if only_new:
        cmd.append("-n")
    subprocess.run(cmd + [folder, bucket])


def bucket_dl(bucket, folder, only_new=True):
    gu = shutil.which("gsutil")
    bucket = gsutil_bucket(bucket)
    bucket = gsutil_src(bucket)
    os.makedirs(folder, exist_ok=True)
    cmd = [gu, "-m", "cp", "-r"]
    if only_new:
        cmd.append("-n")
    subprocess.run(cmd + [bucket, folder])


def get_local_path(folder, task_folder):
    if is_bucket(folder):
        if Path(folder).suffix != "":
            return Path(task_folder)
        return Path(task_folder) / Path(folder).name
    else:
        return Path(folder)


def lit_eval(x):
    try:
        return literal_eval(x)
    except:
        return x


def find_alternate_path(path):
    path = Path(path)
    idx = 0
    file_start = "/".join(path.parts[:-1])
    if file_start[:2] == "//":
        file_start = file_start[1:]
    file_start = Path(file_start)
    file_end = path.stem
    new_path = file_start / f"{file_end}{path.suffix}"
    while new_path.exists():
        new_path = file_start / f"{file_end}_{idx}{path.suffix}"
        idx += 1
    msg.info(f"{path} already exists. Using {new_path} instead.", spaced=True)
    return new_path


def resolve_ds_path(ds_path, append=False, overwrite=False):
    ds_path = Path(ds_path)
    if ds_path.is_dir():
        if append:
            msg.info(f"{ds_path} already exists. Appending because append=True.", spaced=True)
            return ds_path
        elif overwrite:
            msg.info(
                f"\n{ds_path} already exists. Overwriting because overwrite=True.\n",
                spaced=True,
            )
            shutil.rmtree(ds_path)
            return ds_path
        ds_path = find_alternate_path(ds_path)
    return ds_path


def write_ds(ds, ds_path, append=False, overwrite=False, **kwargs):
    ds_path = resolve_ds_path(ds_path, append, overwrite=overwrite)
    ds.write_parquet(ds_path, **kwargs)
    return ds_path


def chain_models(models):
    if not is_list(models):
        models = [models]
    return nn.Sequential(*models)


def is_preprocessor(x):
    return isinstance(x, rd.Preprocessor)


def chain_processors(processors):
    if not is_list(processors):
        processors = [processors]
    return Chain(*processors)


def handle_processors(processors, batch_size=None):
    if processors is None:
        return None
    if not is_list(processors):
        processors = [processors]
    if len(processors) == 0:
        return None

    def to_bm(p, bs):
        if not is_preprocessor(p) and callable(p):
            return BatchMapper(p, batch_size=bs, batch_format="pandas")
        elif is_preprocessor(p):
            return p

    return chain_processors([to_bm(p, batch_size) for p in processors if p is not None])


def repartition_ds(ds, num_blocks=2):
    if path_or_str(ds):
        ds = rd.read_parquet(ds)
    try:
        if num_blocks is not None and num_blocks > 0 and ds.num_blocks() != num_blocks:
            ds = ds.repartition(num_blocks)
    except:
        pass
    return ds


def transform_ds(ds, processors=[], num_blocks=2, batch_size=32, **kwargs):
    ds = repartition_ds(ds, num_blocks=num_blocks)
    pp = handle_processors(processors, batch_size=batch_size)
    if pp is None:
        return ds
    return pp.transform(ds)


def group_df_on(df, group_on="path", agg_on=["text"]):
    if not is_list(agg_on):
        agg_on = [agg_on]
    agg_dict = {k: lambda x: list(x) for k in agg_on}
    return df.groupby(group_on, as_index=False).agg(agg_dict).reset_index(drop=True)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_mapper.ipynb.

# %% auto 0
__all__ = ['Callback', 'msg_bs_cb', 'cbs_before_batch', 'cbs_before_batch_rows', 'cbs_after_batch_rows', 'cbs_after_batch',
           'Mapper']

# %% ../nbs/01_mapper.ipynb 2
from .imports import *
from .utils import *


# %% ../nbs/01_mapper.ipynb 4
class Callback:
    "Base class for callbacks for `Mapper`."
    def before_batch(self, **kwargs):
        pass

    def before_batch_rows(self, **kwargs):
        pass

    def after_batch_rows(self, **kwargs):
        pass

    def after_batch(self, **kwargs):
        pass


class msg_bs_cb(Callback):
    "A callback to print the batch size."
    def before_batch(self, df, **kwargs):
        msg.info(f"DF BATCH SIZE: {len(df)}", spaced=True)


def cbs_before_batch(cbs, **kwargs):
    [cb.before_batch(**kwargs) for cb in cbs]


def cbs_before_batch_rows(cbs, df, **kwargs):
    [cb.before_batch_rows(**kwargs) for _ in range(len(df)) for cb in cbs]


def cbs_after_batch_rows(cbs, df, **kwargs):
    [cb.after_batch_rows(**kwargs) for _ in range(len(df)) for cb in cbs]


def cbs_after_batch(cbs, **kwargs):
    [cb.after_batch(**kwargs) for cb in cbs]


class Mapper:
    """
    A class to map a function to a dataframe.
    The function should expect a dataframe as input and return a dataframe as output.
    """

    def __init__(
        self,
        udf=noop,  # A function that takes a dataframe as input and returns a dataframe as output.
        udf_kwargs={},  # The keyword arguments to pass to the `udf`.
        cbs=[],  # A list of `Callback`s to run before and after the mapping. It will add the `msg_bs_cb` by default.
        **kwargs,
    ):
        cbs = [msg_bs_cb()] + cbs
        udf = partial(udf, **udf_kwargs)
        store_attr(**locals_to_params(locals()))

    def map(self, df):
        return df.apply(self.udf, axis=1, result_type="expand")

    def __call__(self, df):
        cbs_before_batch(self.cbs, df=df)
        cbs_before_batch_rows(self.cbs, df=df)

        df = self.map(df)

        cbs_after_batch_rows(self.cbs, df=df)
        cbs_after_batch(self.cbs, df=df)

        return df




# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/dreamai_ray',
                'doc_host': 'https://HamzaFarhan.github.io',
                'git_url': 'https://github.com/HamzaFarhan/dreamai_ray',
                'lib_path': 'dreamai_ray'},
  'syms': { 'dreamai_ray.imports': {},
            'dreamai_ray.mapper': { 'dreamai_ray.mapper.Callback': ('mapper.html#callback', 'dreamai_ray/mapper.py'),
                                    'dreamai_ray.mapper.Callback.after_batch': ( 'mapper.html#callback.after_batch',
                                                                                 'dreamai_ray/mapper.py'),
                                    'dreamai_ray.mapper.Callback.after_batch_rows': ( 'mapper.html#callback.after_batch_rows',
                                                                                      'dreamai_ray/mapper.py'),
                                    'dreamai_ray.mapper.Callback.before_batch': ( 'mapper.html#callback.before_batch',
                                                                                  'dreamai_ray/mapper.py'),
                                    'dreamai_ray.mapper.Callback.before_batch_rows': ( 'mapper.html#callback.before_batch_rows',
                                                                                       'dreamai_ray/mapper.py'),
                                    'dreamai_ray.mapper.Mapper': ('mapper.html#mapper', 'dreamai_ray/mapper.py'),
                                    'dreamai_ray.mapper.Mapper.__call__': ('mapper.html#mapper.__call__', 'dreamai_ray/mapper.py'),
                                    'dreamai_ray.mapper.Mapper.__init__': ('mapper.html#mapper.__init__', 'dreamai_ray/mapper.py'),
                                    'dreamai_ray.mapper.Mapper.map': ('mapper.html#mapper.map', 'dreamai_ray/mapper.py'),
                                    'dreamai_ray.mapper.cbs_after_batch': ('mapper.html#cbs_after_batch', 'dreamai_ray/mapper.py'),
                                    'dreamai_ray.mapper.cbs_after_batch_rows': ( 'mapper.html#cbs_after_batch_rows',
                                                                                 'dreamai_ray/mapper.py'),
                                    'dreamai_ray.mapper.cbs_before_batch': ('mapper.html#cbs_before_batch', 'dreamai_ray/mapper.py'),
                                    'dreamai_ray.mapper.cbs_before_batch_rows': ( 'mapper.html#cbs_before_batch_rows',
                                                                                  'dreamai_ray/mapper.py'),
                                    'dreamai_ray.mapper.msg_bs_cb': ('mapper.html#msg_bs_cb', 'dreamai_ray/mapper.py'),
                                    'dreamai_ray.mapper.msg_bs_cb.before_batch': ( 'mapper.html#msg_bs_cb.before_batch',
                                                                                   'dreamai_ray/mapper.py')},
            'dreamai_ray.pdf.app': { 'dreamai_ray.pdf.app.EmsPredictor': ('pdf/app.html#emspredictor', 'dreamai_ray/pdf/app.py'),
                                     'dreamai_ray.pdf.app.EmsPredictor.__call__': ( 'pdf/app.html#emspredictor.__call__',
                                                                                    'dreamai_ray/pdf/app.py'),
                                     'dreamai_ray.pdf.app.EmsPredictor.__init__': ( 'pdf/app.html#emspredictor.__init__',
                                                                                    'dreamai_ray/pdf/app.py'),
                                     'dreamai_ray.pdf.app.NERPredictor': ('pdf/app.html#nerpredictor', 'dreamai_ray/pdf/app.py'),
                                     'dreamai_ray.pdf.app.NERPredictor.__call__': ( 'pdf/app.html#nerpredictor.__call__',
                                                                                    'dreamai_ray/pdf/app.py'),
                                     'dreamai_ray.pdf.app.NERPredictor.__init__': ( 'pdf/app.html#nerpredictor.__init__',
                                                                                    'dreamai_ray/pdf/app.py'),
                                     'dreamai_ray.pdf.app.SegsPredictor': ('pdf/app.html#segspredictor', 'dreamai_ray/pdf/app.py'),
                                     'dreamai_ray.pdf.app.SegsPredictor.__call__': ( 'pdf/app.html#segspredictor.__call__',
                                                                                     'dreamai_ray/pdf/app.py'),
                                     'dreamai_ray.pdf.app.SegsPredictor.__init__': ( 'pdf/app.html#segspredictor.__init__',
                                                                                     'dreamai_ray/pdf/app.py'),
                                     'dreamai_ray.pdf.app.ner_pipeline': ('pdf/app.html#ner_pipeline', 'dreamai_ray/pdf/app.py'),
                                     'dreamai_ray.pdf.app.ner_pipeline_params': ( 'pdf/app.html#ner_pipeline_params',
                                                                                  'dreamai_ray/pdf/app.py'),
                                     'dreamai_ray.pdf.app.pdf_pipeline': ('pdf/app.html#pdf_pipeline', 'dreamai_ray/pdf/app.py'),
                                     'dreamai_ray.pdf.app.pdf_pipeline_params': ( 'pdf/app.html#pdf_pipeline_params',
                                                                                  'dreamai_ray/pdf/app.py')},
            'dreamai_ray.pdf.core': { 'dreamai_ray.pdf.core.cid_to_char': ('pdf/core.html#cid_to_char', 'dreamai_ray/pdf/core.py'),
                                      'dreamai_ray.pdf.core.extract_text': ('pdf/core.html#extract_text', 'dreamai_ray/pdf/core.py'),
                                      'dreamai_ray.pdf.core.extract_text_dict': ( 'pdf/core.html#extract_text_dict',
                                                                                  'dreamai_ray/pdf/core.py'),
                                      'dreamai_ray.pdf.core.extract_text_list': ( 'pdf/core.html#extract_text_list',
                                                                                  'dreamai_ray/pdf/core.py'),
                                      'dreamai_ray.pdf.core.process_line': ('pdf/core.html#process_line', 'dreamai_ray/pdf/core.py'),
                                      'dreamai_ray.pdf.core.process_text': ('pdf/core.html#process_text', 'dreamai_ray/pdf/core.py')},
            'dreamai_ray.pdf.df': { 'dreamai_ray.pdf.df.create_ner_df': ('pdf/df.html#create_ner_df', 'dreamai_ray/pdf/df.py'),
                                    'dreamai_ray.pdf.df.create_pdf_df': ('pdf/df.html#create_pdf_df', 'dreamai_ray/pdf/df.py'),
                                    'dreamai_ray.pdf.df.df_segs': ('pdf/df.html#df_segs', 'dreamai_ray/pdf/df.py'),
                                    'dreamai_ray.pdf.df.extract_df_text': ('pdf/df.html#extract_df_text', 'dreamai_ray/pdf/df.py'),
                                    'dreamai_ray.pdf.df.write_df_ems': ('pdf/df.html#write_df_ems', 'dreamai_ray/pdf/df.py'),
                                    'dreamai_ray.pdf.df.write_df_ner': ('pdf/df.html#write_df_ner', 'dreamai_ray/pdf/df.py'),
                                    'dreamai_ray.pdf.df.write_df_segs': ('pdf/df.html#write_df_segs', 'dreamai_ray/pdf/df.py')},
            'dreamai_ray.pdf.extract': { 'dreamai_ray.pdf.extract.add_other_class': ( 'pdf/extract.html#add_other_class',
                                                                                      'dreamai_ray/pdf/extract.py'),
                                         'dreamai_ray.pdf.extract.load_ems_model': ( 'pdf/extract.html#load_ems_model',
                                                                                     'dreamai_ray/pdf/extract.py'),
                                         'dreamai_ray.pdf.extract.load_segs_model': ( 'pdf/extract.html#load_segs_model',
                                                                                      'dreamai_ray/pdf/extract.py'),
                                         'dreamai_ray.pdf.extract.print_segments': ( 'pdf/extract.html#print_segments',
                                                                                     'dreamai_ray/pdf/extract.py'),
                                         'dreamai_ray.pdf.extract.text_to_ems': ( 'pdf/extract.html#text_to_ems',
                                                                                  'dreamai_ray/pdf/extract.py'),
                                         'dreamai_ray.pdf.extract.text_to_segments': ( 'pdf/extract.html#text_to_segments',
                                                                                       'dreamai_ray/pdf/extract.py'),
                                         'dreamai_ray.pdf.extract.write_embeddings': ( 'pdf/extract.html#write_embeddings',
                                                                                       'dreamai_ray/pdf/extract.py'),
                                         'dreamai_ray.pdf.extract.write_segments': ( 'pdf/extract.html#write_segments',
                                                                                     'dreamai_ray/pdf/extract.py')},
            'dreamai_ray.pdf.mappers': { 'dreamai_ray.pdf.mappers.SegsMapper': ( 'pdf/mappers.html#segsmapper',
                                                                                 'dreamai_ray/pdf/mappers.py'),
                                         'dreamai_ray.pdf.mappers.SegsMapper.__init__': ( 'pdf/mappers.html#segsmapper.__init__',
                                                                                          'dreamai_ray/pdf/mappers.py')},
            'dreamai_ray.pdf.ner': { 'dreamai_ray.pdf.ner.edu_ner': ('pdf/ner.html#edu_ner', 'dreamai_ray/pdf/ner.py'),
                                     'dreamai_ray.pdf.ner.is_valid_jner': ('pdf/ner.html#is_valid_jner', 'dreamai_ray/pdf/ner.py'),
                                     'dreamai_ray.pdf.ner.is_valid_tner': ('pdf/ner.html#is_valid_tner', 'dreamai_ray/pdf/ner.py'),
                                     'dreamai_ray.pdf.ner.job_ner': ('pdf/ner.html#job_ner', 'dreamai_ray/pdf/ner.py'),
                                     'dreamai_ray.pdf.ner.load_job_model': ('pdf/ner.html#load_job_model', 'dreamai_ray/pdf/ner.py'),
                                     'dreamai_ray.pdf.ner.load_ner_model': ('pdf/ner.html#load_ner_model', 'dreamai_ray/pdf/ner.py'),
                                     'dreamai_ray.pdf.ner.proc_ner': ('pdf/ner.html#proc_ner', 'dreamai_ray/pdf/ner.py'),
                                     'dreamai_ray.pdf.ner.work_ner': ('pdf/ner.html#work_ner', 'dreamai_ray/pdf/ner.py')},
            'dreamai_ray.redis_kv_store': {},
            'dreamai_ray.utils': { 'dreamai_ray.utils.bucket_dl': ('utils.html#bucket_dl', 'dreamai_ray/utils.py'),
                                   'dreamai_ray.utils.bucket_move': ('utils.html#bucket_move', 'dreamai_ray/utils.py'),
                                   'dreamai_ray.utils.bucket_up': ('utils.html#bucket_up', 'dreamai_ray/utils.py'),
                                   'dreamai_ray.utils.chain_models': ('utils.html#chain_models', 'dreamai_ray/utils.py'),
                                   'dreamai_ray.utils.chain_processors': ('utils.html#chain_processors', 'dreamai_ray/utils.py'),
                                   'dreamai_ray.utils.find_alternate_path': ('utils.html#find_alternate_path', 'dreamai_ray/utils.py'),
                                   'dreamai_ray.utils.get_local_path': ('utils.html#get_local_path', 'dreamai_ray/utils.py'),
                                   'dreamai_ray.utils.get_task_from_kv_store': ( 'utils.html#get_task_from_kv_store',
                                                                                 'dreamai_ray/utils.py'),
                                   'dreamai_ray.utils.group_df_on': ('utils.html#group_df_on', 'dreamai_ray/utils.py'),
                                   'dreamai_ray.utils.gsutil_bucket': ('utils.html#gsutil_bucket', 'dreamai_ray/utils.py'),
                                   'dreamai_ray.utils.gsutil_src': ('utils.html#gsutil_src', 'dreamai_ray/utils.py'),
                                   'dreamai_ray.utils.handle_processors': ('utils.html#handle_processors', 'dreamai_ray/utils.py'),
                                   'dreamai_ray.utils.init_task_progress': ('utils.html#init_task_progress', 'dreamai_ray/utils.py'),
                                   'dreamai_ray.utils.is_bucket': ('utils.html#is_bucket', 'dreamai_ray/utils.py'),
                                   'dreamai_ray.utils.is_preprocessor': ('utils.html#is_preprocessor', 'dreamai_ray/utils.py'),
                                   'dreamai_ray.utils.json_file': ('utils.html#json_file', 'dreamai_ray/utils.py'),
                                   'dreamai_ray.utils.lit_eval': ('utils.html#lit_eval', 'dreamai_ray/utils.py'),
                                   'dreamai_ray.utils.repartition_ds': ('utils.html#repartition_ds', 'dreamai_ray/utils.py'),
                                   'dreamai_ray.utils.resolve_ds_path': ('utils.html#resolve_ds_path', 'dreamai_ray/utils.py'),
                                   'dreamai_ray.utils.transform_ds': ('utils.html#transform_ds', 'dreamai_ray/utils.py'),
                                   'dreamai_ray.utils.update_task_progress': ('utils.html#update_task_progress', 'dreamai_ray/utils.py'),
                                   'dreamai_ray.utils.write_ds': ('utils.html#write_ds', 'dreamai_ray/utils.py')}}}



import redis
from wasabi import msg


def connect_to_server(redis_host="127.0.0.1", redis_port=6379):
    try:
        r = redis.StrictRedis(
            host=redis_host, port=redis_port, charset="utf-8", decode_responses=True
        )
    except Exception as e:
        raise Exception(f"Connecting to Redis Server failed with error {e}")
    return r


class KeyValueStore:
    def __init__(self, redis_host="127.0.0.1", redis_port=6379):
        try:
            self.server = connect_to_server(redis_host=redis_host, redis_port=redis_port)
            # self.logger = logger
        except Exception as e:
            raise

    def insert(self, key, value):
        server = self.server
        # ogger.debug(key,value)
        msg.info(key, value, spaced=True)
        server.hmset(key, value)

    def get(self, key):
        server = self.server
        # logger.debug(key)
        try:
            val = server.hgetall(key)
        except Exception as e:
            # logger.error("unable to retrieve value of key {} from Redis: error = {}".format(key,e))
            raise f"Unable to retrieve value of key {key} from Redis: error = {e}"
        return val

    def getall(self):
        return self.server.keys()

    def remove(self, key):
        server = self.server
        # logger.debug('removing key {}'.format(key))
        msg.info("removing key {}".format(key), spaced=True)
        try:
            all_keys = list(server.hgetall(key).keys())
            server.hdel(key, *all_keys)
            # logger.debug('key {} removed'.format(key))
            msg.info("key {} removed".format(key), spaced=True)
        except Exception as e:
            # logger.error("unable to remove key {} from Redis: error = {}".format(key,e))
            raise f"Unable to remove key {key} from Redis: error = {e}"



from dreamai.imports import *
from dreamai.core import *

import requests
import subprocess
from torch import nn

# import pandas as pd
from pypdf import PdfReader
from ast import literal_eval
from setfit import SetFitModel
from fastai.torch_core import default_device
from sentence_transformers import SentenceTransformer
from transformers import (
    AutoTokenizer,
    AutoModel,
    AutoModelForSequenceClassification,
    AutoModelForTokenClassification,
    pipeline,
)

import ray
from ray import serve

# import modin.pandas as pd
from ray import data as rd
from ray.air import Checkpoint
from ray.train.predictor import Predictor
from ray.train.batch_predictor import BatchPredictor
from ray.data.preprocessors import Chain, BatchMapper
from ray.serve.air_integrations import PredictorDeployment
from ray.air.util.data_batch_conversion import BatchFormat
from ray.serve.http_adapters import pandas_read_json, json_request



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/pdf/02_df.ipynb.

# %% auto 0
__all__ = ['create_pdf_df', 'create_ner_df', 'extract_df_text', 'write_df_ner', 'df_segs', 'write_df_segs', 'write_df_ems']

# %% ../../nbs/pdf/02_df.ipynb 2
from ..imports import *
from ..utils import *
from .core import *
from .extract import *
from .ner import *


# %% ../../nbs/pdf/02_df.ipynb 4
def create_pdf_df(data_path, segs_folder, ems_folder):
    files = [str(x) for x in resolve_data_path(data_path) if Path(x).suffix == ".pdf"]
    df = pd.DataFrame({"path": files})
    df["segs_folder"] = str(segs_folder)
    df["ems_folder"] = str(ems_folder)
    return df


def create_ner_df(data_path, segs_folder):
    data_path = Path(data_path)
    if data_path.is_file():
        files = [str(data_path)]
    else:
        files = get_files(data_path, extensions=[".pdf"], make_str=True)
    df = pd.DataFrame({"path": files})
    df["segs_folder"] = str(segs_folder)
    return df


def extract_df_text(df, col="path", n_lines=3):
    df["text"] = extract_text_list(df[col], n_lines=n_lines)
    return df


def write_df_ner(df, tner, jner, work_ner_dict={"company": "", "date": ""}):
    segs_path = json_file(df["path"], df["segs_folder"])
    if os.path.exists(segs_path):
        with open(segs_path, "r") as f:
            segs = json.load(f)
        if not is_list(segs["Work Experience"][0]):
            segs = {k: [v] for k, v in segs.items()}
        work_ners = [
            work_ner(txt, tner, jner, ner_dict=work_ner_dict)
            for txt in segs["Work Experience"][0]
        ]
        work_ners = [ner for ner in work_ners if len(ner) > 0]
        segs["Work Experience"].append(work_ners)
        edu_ners = [edu_ner(txt, tner) for txt in segs["Education"][0]]
        edu_ners = [ner for ner in edu_ners if len(ner) > 0]
        segs["Education"].append(edu_ners)
        ner_path = extend_path_name(segs_path, "_ner")
        with open(ner_path, "w") as f:
            json.dump(segs, f, indent=4)
    return df


def df_segs(
    df,
    segs_model,
    thresh=0.6,
    classes=["Work Experience", "Education", "Certifications", "Other"],
    other_class="Other",
    write=False,
):
    if len(df["text"]) == 0:
        return {}, [], np.array([])
    segs, preds, probs = text_to_segments(
        df["text"], segs_model, thresh, classes=classes, other_class=other_class
    )
    if write:
        segs_path = json_file(df["path"], df["segs_folder"])
        with open(segs_path, "w") as f:
            json.dump(segs, f, indent=4)
    df["segs"] = segs
    df["preds"] = preds
    df["probs"] = probs
    return df


def write_df_segs(
    df,
    segs_model,
    thresh=0.6,
    write=False,
    classes=["Work Experience", "Education", "Certifications", "Other"],
    other_class="Other",
):
    if len(df["text"]) == 0:
        return {}, [], np.array([])
    segs, preds, probs = text_to_segments(
        df["text"], segs_model, thresh, classes=classes, other_class=other_class
    )
    if write:
        segs_path = json_file(df["path"], df["segs_folder"])
        with open(segs_path, "w") as f:
            json.dump(segs, f, indent=4)
    return segs, preds, probs


def write_df_ems(
    df,
    ems_model,
    classes=["Work Experience", "Education", "Certifications", "Other"],
    other_class="Other",
):
    classes = add_other_class(classes, other_class)
    seg_ids = {c: i + 1 for i, c in enumerate(classes)}
    fn = Path(df["path"]).stem
    ems_folder = Path(df["ems_folder"])
    os.makedirs(ems_folder, exist_ok=True)
    ems_path = str(ems_folder / fn)
    txt = df["text"]
    pred_classes = df.get("classes", [""] * len(txt))
    if len(txt) == 0 or len(pred_classes) == 0:
        return df
    ems = ems_model.encode(txt).tolist()
    for i, data in enumerate(zip(pred_classes, ems)):
        pc, em = data
        seg_id = seg_ids.get(pc, 0)
        if seg_id == 0:
            seg = ""
        else:
            seg = f"_{seg_id}"
        jf = ems_path + f"{seg}_{i+1}.json"
        em_dict = {"id": fn, "embedding": em}
        with open(jf, "w") as f:
            json.dump(em_dict, f)
    return df
